%{
    open Types
%}

%token <int> INT
%token <string> ID
%token TRUE
%token FALSE
%token PLUS
%token MINUS
%token TIMES
%token EQUALS
%token LT
%token GT
%token AND
%token OR
%token LPAREN
%token RPAREN
%token LET
%token LETREC
%token FN
%token IN
%token IF
%token THEN
%token ELSE
%token FN_ARROW
%token TY_ARROW
%token EOF
%token TYPE
%token INT_TYPE
%token BOOL_TYPE


%nonassoc IN
%nonassoc ELSE
%left LT
%left GT
%left PLUS
%left TIMES
%left OR
%left AND

%start <Exp.t> prog
%%

prog:
| e = expr; EOF { e }
;


ty :
  | INT_TYPE { Int }
  | BOOL_TYPE { Bool }
  | t1 = ty; TY_ARROW; t2 = ty { Exp.Arrow (t1,t2) }
  | t1 = ty; TIMES; t2 = ty { Exp.Tuple([t1;t2]) }
  | LPAREN; t = ty; RPAREN { t }
;      



expr :
  | i = INT { Const (I i) }
  | TRUE { Const (B true) }
  | FALSE { Const (B false) }
  | x = ID { Var x }
  | FN; LPAREN; arg = separated_pair(ID,TYPE,ty); RPAREN; TYPE; rettype = ty; FN_ARROW; body = expr { Const (TyFn (arg,rettype,body)) }
  | FN; LPAREN; arg = ID; RPAREN; FN_ARROW; body = expr { Const (Fn (arg,body)) }
  | e1 = expr; PLUS;  e2 = expr { PrimCall(Plus,[e1;e2]) }
  | e1 = expr; MINUS; e2 = expr { PrimCall(Minus,[e1;e2]) }
  | e1 = expr; TIMES; e2 = expr { PrimCall(Times,[e1;e2]) }
  | e1 = expr; LT; e2 = expr { PrimCall(Lt,[e1;e2]) }
  | e1 = expr; GT; e2 = expr { PrimCall(Gt,[e1;e2]) }
  | e1 = expr; AND; e2 = expr { PrimCall(And,[e1;e2]) }
  | e1 = expr; OR; e2 = expr { PrimCall(Or,[e1;e2]) }
  | e1 = expr; EQUALS; e2 = expr { PrimCall(Equal,[e1;e2]) }
  | LET; x = ID; TYPE; t1 = ty; EQUALS; e1 = expr; IN; e2 = expr { TyBind((x,t1),e1,e2) }
  | LET; x = ID; EQUALS; e1 = expr; IN; e2 = expr { Bind(x,e1,e2) }
  | LETREC; x = ID; LPAREN; arg = separated_pair(ID,TYPE,ty); RPAREN; TYPE; rettype = ty; FN_ARROW; body = expr; IN; e = expr { TyBindRec(x,arg,rettype,body,e) }
  | LETREC; x = ID; LPAREN; arg = ID; RPAREN; FN_ARROW; body = expr; IN; e = expr { BindRec(x,arg,body,e) }
  | x = expr; arg = expr; { Call(x,arg) }
  | IF; e1 = expr; THEN; e2 = expr; ELSE; e3 = expr { If(e1,e2,e3) }
  | LPAREN; e = expr; RPAREN {e}
;

